/* 
 * twio.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "twin.h"

#define BUFFER_LENGTH 1024

int _far _cdecl
twprintf(const char _far * format, ...)
{
    char buffer[BUFFER_LENGTH + 1];
    int result;
    va_list arg_ptr;

    va_start(arg_ptr, format);
    if((result = vsprintf(buffer, format, arg_ptr)) > 0){
        buffer[BUFFER_LENGTH] = '\0';
	wputs(buffer, GetHomeWnd());
    }
    va_end(arg_ptr);
    return result;
} 

int _far _cdecl 
twfprintf(FILE _far * fp, const char _far * format, ...)
{
    char buffer[BUFFER_LENGTH + 1];
    int result;
    va_list arg_ptr;

    va_start(arg_ptr, format);
    if (fp == stdin) {
    	result = -1;
    } else if(fp == stdout || fp == stderr) {  
    	result = vsprintf(buffer, format, arg_ptr);
	wputs(buffer, GetHomeWnd());
    } else {
	result =vfprintf(fp, format, arg_ptr);
    }
    va_end(arg_ptr);
    return result;
}

int _far _cdecl
twfputc(char ch, FILE _far * fp)
{
    if (fp == stdin) {
        return EOF;
    } else if (fp == stdout || fp == stderr) {
	WriteStreem(&ch, 1);
	return ch;
    }
    return fputc(ch, fp);
}

int _far _cdecl
twfputs(const char _far * string, FILE _far * fp)
{
    if (fp == stdin) {
	return 0;
    } else if (fp == stdout || fp == stderr) {
    	wputs(string,  GetHomeWnd());
	return 1;
    }
    return fputs(string, fp);
}

void _far _cdecl
twclearerr(FILE _far * fp)
{
    if (fp == stdin || fp == stdout || fp == stderr) {
	return;
    }
    clearerr(fp);
}

      
int _far _cdecl
twfileno(FILE _far * fp)		 
{
    if (fp == stdin || fp == stdout || fp == stderr) {
	return (int) (long) fp;
    }
    return fileno(fp);
}

int _far _cdecl
twferror(FILE _far * fp)
{
    if (fp == stdin || fp == stdout || fp == stderr) {
	return 0;
    }
    return ferror(fp);
}

int _far _cdecl
twfeof(FILE _far * fp)
{
    if (fp == stdin || fp == stdout || fp == stderr) {
	return 0;
    }
    return feof(fp);
}

int _far _cdecl
twfflush(FILE _far * fp)
{
    if (fp == stdin) {
	return FlushStreem();
    } else if (fp == stdout || fp == stderr) {
	return 0;
    }
    return fflush(fp);
}

char _far * _far _cdecl
twfgets(char _far * buffer, int size, FILE _far * fp)
{
    if (fp == stdin) {
        char *bp;
        char ch;
        int cnt;
       
        for (cnt = size - 1, bp = buffer; cnt > 0; cnt--, bp++) {
            if (ReadStreem(&ch, 1) <= 0) {
                return NULL;
            } else if (ch == '\n') {
                *bp = '\n';
		bp++;
		break;
            } else {
                *bp = ch;
            }
         }
	 *bp = '\0';
         return buffer;
     } else if (fp == stdout || fp == stderr) {
         return NULL;
     } else {
         return fgets(buffer, size, fp);
     }
}

int _far _cdecl
twfgetc(FILE _far * fp)
{
    if (fp == stdin) {
        char ch;
        if (ReadStreem(&ch, 1) <= 0) {
           return EOF;
        }
        return (int) ch;
    } else if (fp == stdout || fp == stderr) {
	return EOF;
    } else {
    	return fgetc(fp);
    }
}
