/* 
 * twlib.c
 *
 * Copyright (c) 1994,1995 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include "twin.h"
#include <unistd.h>
#include <direct.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/timeb.h>
#ifdef KANJI
#include <mbstring.h>
#endif /* KANJI */

struct timeval {
    long tv_sec;
    long tv_usec;
};

struct timezone {
    int tz_minuteswet;
    int tz_dsttime;
};

extern void gettimeofday(struct timeval *, struct timezone *);

int _cdecl
pipe()
{
    return -1;
}
int _cdecl
fork()
{
    return -1;
}
int _cdecl
geteuid()
{
    return 666;
}
int _cdecl
_execvp()
{
    return -1;
}
void _cdecl
_exit(int val)
{
    jmp_buf env;
    memcpy(env, GetJmpBuf(), sizeof(jmp_buf)); 
    longjmp(env, val + 1);
}

void _cdecl
exit(int val)
{
    jmp_buf env;
    memcpy(env, GetJmpBuf(), sizeof(jmp_buf)); 
    longjmp(env, val + 1);
}

#undef waitpid
int _cdecl
waitpid()
{
    return -1;
}

char * _cdecl
getcwd(buffer, size)
    char *buffer;
    int size;
{
    int length;
    char *pb;

    if (_getcwd(buffer, size) == NULL) {
	errno = EACCES;
	return NULL;
    }
    length = strlen(buffer);
    if (length >= size) {
	errno = ERANGE;
	return NULL;
    }

    strcpy(buffer, strchr(buffer, '\\'));
#ifdef KANJI    
    _mbslwr(buffer);
#else    
    strlwr(buffer);
#endif /* KANJI */
    for (pb = buffer; *pb != '\0'; pb++) {
    	if (*pb == '\\')
	    *pb = '/';
    }
    return buffer;
}

void _cdecl
closedir(DIR *dirp)
{
    free(dirp);
}

DIR * _cdecl
opendir(char *path)
{
    char Buffer[_MAX_PATH+1];
    DIR *dd;
    char *pb;
    int len;
        
    if ((len = strlen(path)) > _MAX_PATH - 4) {
	errno = ERANGE;
        return NULL;
    }

    for (pb = Buffer; *path != '\0'; pb++) {
	if ((*pb = *path) == '/')
	    while (*path == '/') path++;
	else
	    path++;
    }
    if (*(pb-1) != '/')	{
    	*pb = '/';
	pb++;
    }
    strcpy(pb, "*.*"); 
    	
    dd = (DIR *) malloc(sizeof(DIR));
    if (!_dos_findfirst(Buffer, _A_SUBDIR, &dd->dd_info)) {
	dd->dd_flag = 1;
        return dd;
    }
    return NULL;
}

struct dirent *	_cdecl
readdir(DIR *dirp)
{                    
    if (dirp->dd_flag == 0) 
    	return NULL;
        
    strcpy(dirp->dd_entry.d_name, dirp->dd_info.name);
#ifdef KANJI    
    _mbslwr(dirp->dd_entry.d_name);
#else    
    strlwr(dirp->dd_entry.d_name);
#endif /* KANJI */
    dirp->dd_entry.d_namlen = strlen(dirp->dd_info.name);
    if (_dos_findnext(&dirp->dd_info))
	dirp->dd_flag = 0;

    return &dirp->dd_entry;
}

struct passwd * _cdecl
getpwnam(char *name)
{
    return NULL;
}

void _cdecl
endpwent(void)
{
}

void _cdecl
gettimeofday(struct timeval *tv, struct timezone *tz)
{                                
    struct timeb tb;

    ftime(&tb);
    tv->tv_sec = tb.time;
    tv->tv_usec = (long) tb.millitm * 1000;
}
