/*
 * twmodule.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#define OpenFile OPenFile
#include <windows.h>
#undef 	OpenFile
#include "tclInt.h"
#include "tclUnix.h"
//#define SW_HIDE             0
//#define SW_SHOWMINIMIZED    2
//#define SW_SHOWMAXIMIZED    3
//#define SW_SHOW             5

static char *dtStr[] = {
	"DT_PLOTTER",
	"DT_RASDISPLAY",
	"DT_RASPRINTER",
	"DT_RASCAMERA",
	"DT_CHARSTREAM",
	"DT_METAFILE",
	"DT_DISPFILE"      
};

static int initiarized = 0;

Tcl_HashTable libraryTable;

typedef struct {
    unsigned hinst;
    int cont;
} Library;


/*
 *----------------------------------------------------------------------
 *
 * Twin_WinExecCmd --
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
FAR PASCAL
Twin_WinExecCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    unsigned wReturn;
    char * lpszCmdLine;
    unsigned fuCmdShow = SW_SHOW;
    int index = 1;
    char buffer[32];

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], "?-cmdShow? arg ?arg ...?", "\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (argv[1][0] == '-') {
	if (strcmp(argv[1], "-show") == 0) {
	} else if (strcmp(argv[1], "-show") == 0) {
	    fuCmdShow = SW_SHOW;
	} else if (strcmp(argv[1], "-minimize") == 0) {
	    fuCmdShow = SW_SHOWMINIMIZED;
	} else if (strcmp(argv[1], "-maximize") == 0) {
	    fuCmdShow = SW_SHOWMAXIMIZED;
	} else if (strcmp(argv[1], "-hide") == 0) {
	    fuCmdShow = SW_HIDE;
	} else {
	    Tcl_AppendResult(interp, "wromg # flags: \"", argv[1],
	    	     "\"", (char *) NULL);
	    return TCL_ERROR;
	}
	index = 2;
    }

    lpszCmdLine = Tcl_Concat(argc - index, &argv[index]);

    wReturn = WinExec(lpszCmdLine, fuCmdShow);
    sprintf(buffer, "%d", wReturn);

    if (wReturn < 32) {
	Tcl_AppendResult(interp, "error getting winexec : ",
		buffer, (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_AppendResult(interp, buffer, (char *)NULL);
    return TCL_OK;
}


void
FAR PASCAL EXPORT
Twin_CallExit(void)
{
    Tcl_HashSearch search;	
    Tcl_HashEntry *hashPtr;
    Library *libPtr;

    hashPtr = Tcl_FirstHashEntry(&libraryTable, &search);
    if (hashPtr == NULL) return;
    do {
        libPtr = (Library *) Tcl_GetHashValue(hashPtr);
	FreeLibrary(libPtr->hinst);
	free(libPtr);
	Tcl_DeleteHashEntry(hashPtr);
    } while (hashPtr = Tcl_NextHashEntry(&search));

    Tcl_DeleteHashTable(&libraryTable);
}


/* The code below is quoted from the Tkwin source file msdos.c
 * and is slightly modified.
 *
 *				-- haga@sra.co.jp
 */

char *to_slash(string)
char *string;
{
	static char strn[129];
	char *str;

	str=strn;
	strcpy(strn,string);
	while(*str)
	{
		if(*str=='\\') *str='/';
		str++;
	}
	return(strn);
}

char *to_bslash(string)
char *string;
{
	static char strn[129];
	char *str;

	str=strn;
	strcpy(strn,string);
	while(*str)
	{
		if(*str=='/') *str='\\';
		str++;
	}
	return(strn);
}
/*
 *----------------------------------------------------------------------
 *
 * Tcl_ExecCmd --
 *
 *	This procedure is invoked to process the "exec" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
FAR PASCAL
Tcl_ExecCmd(dummy, interp, argc, argv)
	ClientData dummy;			/* Not used. */
	Tcl_Interp *interp;			/* Current interpreter. */
	int argc;				/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
	FILE *fp=NULL, *fp2 = NULL;
	int i,result,got_pipe=0;
	char command[200]="",tempstr[10];
	char input[40]="",output[40]="",file[32],file2[32],file3[32];
	enum {none,got_com,pipe} status;
	static char tmpdir[32];

	if(!*tmpdir){
		char * td;
		if((td = getenv("TEMP")) == NULL){
		   Tcl_AppendResult( interp, "TEMP env variable not set", NULL);
		   return(TCL_ERROR);

		} else strcpy(tmpdir, td);
	}

	status=none;


	for(i=1;i<argc;i++)
	{
		if(!strcmp(argv[i],"<"))
		{
			if(*input)
			{
			   Tcl_AppendResult( interp, "only one stdin source allowed", NULL );
			   return(TCL_ERROR);

			}
			if(i==argc-1)
			{
			   Tcl_AppendResult( interp, "can't specify \"<\" as last word in command", NULL );
			   return(TCL_ERROR);

			}
			i++;
			sprintf(input," <%s",argv[i]);
		}
		else
		if(!strcmp(argv[i],"<<"))
		{
			if(*input)
			{
			   Tcl_AppendResult( interp, "only one stdin source allowed", NULL );
			   return(TCL_ERROR);

			}
			if(i==argc-1)
			{
			   Tcl_AppendResult( interp, "can't specify \"<<\" as last word in command", NULL );
			   return(TCL_ERROR);

			}
			i++;
			sprintf(file, "%s\\TXXXXXX", tmpdir);
			fp=fopen(mktemp(file),"wt");
			if(!fp) {
				Tcl_AppendResult(interp,
				    "couldn't open temporary file: ",
				    Tcl_PosixError(interp), (char *) NULL);
				return(TCL_ERROR);
			}
			fputs(argv[i],fp);
			fclose(fp);
			sprintf(input," <%s",file);
		}
		else
		if(!strcmp(argv[i],">") || !strcmp(argv[i],">>"))
		{
			if(*output)
			{
			   Tcl_AppendResult( interp, "only one stdout source allowed", NULL );
			   return(TCL_ERROR);

			}
			if(i==argc-1)
			{
			   Tcl_AppendResult( interp, "no \">\" or \">>\" allowed as last word in command", NULL );
			   return(TCL_ERROR);

			}
			i++;
			sprintf(output," %s%s",argv[i-1],argv[i]);
		}
	}
	for(i=1;i<argc;i++)
	{
		if(!strcmp(argv[i],"|"))
		{
			if(status==pipe || status==none || i==argc-1)
			{
			   Tcl_AppendResult( interp, "illegal use of | in command", NULL );
			   return(TCL_ERROR);

			}
			strcat(command,to_bslash(input));
			strcat(command," |");
			status=pipe;
			got_pipe=1;
		}
		else
		if(!strcmp(argv[i],"&")); /* ignore & for background processing */
		else
		if(!strcmp(argv[i],">") || !strcmp(argv[i],">>") ||
		   !strcmp(argv[i],"<") || !strcmp(argv[i],"<<")) i++;
		else
		{
    			Tcl_DString buffer;
			if(i > 1) strcat(command," ");
			strcat(command,
			    to_bslash(Tcl_TildeSubst(interp,argv[i], &buffer)));
			Tcl_DStringFree(&buffer);
			status=got_com;
		}
	}

	if(!strlen(command))
	{
	   Tcl_AppendResult( interp, "didn't specify command to execute", NULL );
	   return(TCL_ERROR);
	}

	if(!*output)
	{
		sprintf(file2, "%s\\TXXXXXX", tmpdir);
		mktemp(file2);
		sprintf(output," >%s",file2);
	}
	else strcpy(file2,"");
	sprintf(file3, "%s\\tcltmp.bat",tmpdir);
	if((fp2=fopen(file3,"wt")) == NULL) {
		Tcl_AppendResult(interp, "couldn't open temporary file: ",
			file3, ": ", Tcl_PosixError(interp), (char *) NULL);
		return(TCL_ERROR);
	}
	fprintf(fp2, "%s", command);
	if(!got_pipe) fprintf(fp2, "%s", to_bslash(input));
	fprintf(fp2, "%s\n", to_bslash(output));
	fclose(fp2);
	
#ifdef NOTDEF
	if(!got_pipe) strcat(command,to_bslash(input));
	strcat(command,to_bslash(output));

	result = system( command );
#endif

	{
	    char cmdBuf[128];
	    sprintf(cmdBuf, "dosprmpt.pif /c %s", file3);
	    result = WinExec(cmdBuf,SW_HIDE);
	    if(result >= 32){
		do Yield();
		while (GetModuleFileName( result,
			cmdBuf, sizeof(cmdBuf)) != 0);
	    }
	}
	if(fp) unlink(file);
	if(fp2) unlink(file3);

	sprintf( tempstr, "%d", result );
	if( result < 32) {
		Tcl_SetErrorCode( interp, "CHILDSTATUS", "0", tempstr, NULL );
		Tcl_AppendResult( interp, "command returned error code: ", tempstr, " during \"", *argv, "\" command", ". Command String: ", command, NULL );

		if(*file2) unlink(file2);

		return(TCL_ERROR);
        } else result = TCL_OK;

       /*
	* Read the child's output (if any) and put it into the result.
	*/

	if(*file2)
	{
		fp=fopen(file2,"rt");
		if(fp)
		{
			while (1)
			{
#	    		define BUFFER_SIZE 511
				char buffer[BUFFER_SIZE+1];
				int count;

				count = fread(buffer,1, BUFFER_SIZE,fp);

				if (count == 0) break;
				if (count < 0)
				{
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp,
					"error reading output: ",
					Tcl_PosixError(interp), (char *) NULL);
					fclose(fp);
					return(TCL_ERROR);
				}
				buffer[count] = 0;
#ifdef NOTDEF
				if(buffer[count-1]=='\n') buffer[count-1]=0;
#endif
				Tcl_AppendResult(interp, buffer, (char *) NULL);
			}
		}
		else
		{
			Tcl_AppendResult( interp, "cannot read input from child"
				, (char *) NULL );
			return(TCL_ERROR);
		}
		fclose(fp);
		if(*file2) unlink(file2);
	}

	return(TCL_OK);
}
                                                                                                                                    