/*
 * tkCtext.c --
 *
 *	This file contains conversion functions between
 *	wchar and STRING/COMPOUND_TEXT encoding.
 *
 * Copyright 1988,1993 Software Research Associates, Inc.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#ifndef lint
static char rcsid[] = "$Header: /mnt/X11/cvsroot/tk/tkCtext.c,v 1.7 1993/08/24 02:06:24 ishisone Exp $";
#endif

#ifdef KANJI

#include "tkConfig.h"
#include "tkInt.h"
#include <mbctype.h>

#pragma optimize("lncg", on)
/*
 *--------------------------------------------------------------
 *
 * Tk_WStrToString --
 *
 *	Convert wchar string to STRING encoding string.
 *	Any characters which cannot be converted are ignored.
 *
 * Results:
 *	The return value is a pointer to the converted string,
 *	or NULL if an error occurred in the conversion process.
 *	The storage for the converted string is allocated with
 *	malloc (ckalloc, precisely), and it is the caller's
 *	responsibility to free it.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

char *
FAR PASCAL
Tk_WStrToString(ws, n)
wchar *ws;
int n;
{
    int len;
    char *s;
    int i, j;

    if (n < 0) {
	wchar *t = ws;
	n = 0;
	while (*t++ != 0) n++;
    }

    for (i = 0, len = 0; i < n; i++) {
	if ((ws[i] & 0xff00) == 0) len++;
	else len += 2;
    }

    if (len <= 0) return NULL;
    s = ckalloc((unsigned int)len + 1);

    for (i = 0, j = 0; i < n; i++) {
	if ((ws[i] & 0xff00) == 0) s[j++] = (char) ws[i];
	else {
	    s[j++] = (char) (ws[i] >> 8);
	    s[j++] = (char ) (ws[i] & 0xff);
	}
    }
    s[j] = '\0';

    return s;
}

/*
 *--------------------------------------------------------------
 *
 * Tk_WStrToCtext --
 *
 *	Convert wchar string to Compound Text string.
 *
 * Results:
 *	The return value is a pointer to the converted string,
 *	or NULL if an error occurred in the conversion process.
 *	The storage for the converted string is allocated with
 *	malloc (ckalloc, precisely), and it is the caller's
 *	responsibility to free it.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

char *
FAR PASCAL
Tk_WStrToCtext(ws, n)
wchar *ws;
int n;
{
    int len;
    char *s;
    int i, j;

    if (n < 0) {
	wchar *t = ws;
	n = 0;
	while (*t++ != 0) n++;
    }

    for (i = 0, len = 0; i < n; i++) {
	if ((ws[i] & 0xff00) == 0) len++;
	else len += 2;
    }

    if (len <= 0) return NULL;
    s = ckalloc((unsigned int)len + 1);

    for (i = 0, j = 0; i < n; i++) {
	if ((ws[i] & 0xff00) == 0) s[j++] = (char) ws[i];
	else {
	    s[j++] = (char) (ws[i] >> 8);
	    s[j++] = (char) (ws[i] & 0xff);
	}
    }
    s[j] = '\0';

    return s;

}

/*
 *--------------------------------------------------------------
 *
 * Tk_CtextToWStr --
 *
 *	Convert Compound Text string to wchar string.  Any
 *	characters which cannot be converted are ignored.
 *	Note that this function can be used to convert STRING
 *	to wchar string, for Compound Text is a superset of
 *	STRING.
 *
 * Results:
 *	The return value is a pointer to the converted string,
 *	or NULL if an error occurred in the conversion process.
 *	The storage for the converted string is allocated with
 *	malloc (ckalloc, precisely), and it is the caller's
 *	responsibility to free it.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

wchar *
FAR PASCAL
Tk_CtextToWStr(ct, n)
char *ct;
int n;
{
    int len;
    wchar *ws;
    int c1, c2;
    int i, j;

    if (n < 0) {
	char *t = ct;
	n = 0;
	while (*t++ != 0) n++;
    }

    for (i = 0, len = 0; i < n; len++ ) {
	if(_ismbblead((unsigned char) ct[i++])) i++;
    }

    if (len <= 0) return NULL;
    ws = ckalloc((unsigned int)(len + 1) * 2);

    for (i = 0, j = 0; i < n; j++) {
        c1 = (unsigned char) ct[i++];
	if( _ismbblead(c1) ) {
	    c2 = (unsigned char) ct[i++];
	    if (ws) ws[j] = (c1 << 8) | c2;
	} else {
	    if (ws) ws[j] = c1;
	}
    }

    if (ws) ws[j] = 0;

    return ws;
}
#endif /* KANJI */


