/*
 * tkWinUtil.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include <windows.h>
#include "tkInt.h"
#include "tkConfig.h"
#include "twin.h"

/*
 * use management color
 */
extern int pure;

static int builded = 0;

static Tcl_HashTable ceDBTable;

typedef struct {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    char *name;
} ColorEntry;

static ColorEntry colorDB[] = {
#include "xcolor2.h"
      0,   0,   0, NULL
};

/*
 * use management cursor 
 */
#define HDR_SIZE 0x46
#define	BM_SIZE 0x80

typedef struct {
	char dummy1[0x0a];
	UINT x_hot, y_hot;
	char dummy2[0x38];
} CURSORFILEHEADER;		

/*
 * use maneagement selection 
 */

typedef struct {
    char *string;		/* Contents of selection are					      
				 * here.  This space is malloc-ed. */
    int bytesAvl;		/* Total number of bytes available
				 * at string. */
    int bytesUsed;		/* Bytes currently in use in string,
				 * not including the terminating
				 * NULL. */
} GetInfo;


extern int	__far __pascal SelGetProc _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, char *portion));

extern void	__far __pascal __export Twin_SetClipboard _ANSI_ARGS_((Tk_Window tkwin));
extern int	__far __pascal __export Twin_GetClipboardData _ANSI_ARGS_((Tcl_Interp *interp, 
			    Tk_GetSelProc *proc, ClientData clientData));

/*
 * BuildColorDB
 *
 */

void
BuildColorDB()
{
    ColorEntry *ePtr;
    Tcl_HashEntry *hPtr;
    int dummy;

    builded = 1;
	Tcl_InitHashTable(&ceDBTable, TCL_STRING_KEYS);
    for (ePtr = colorDB; ePtr->name != NULL; ePtr++) {
	char * p;

	p = malloc(strlen(ePtr->name) + 1);
	strcpy(p, ePtr->name);
        hPtr = Tcl_CreateHashEntry(&ceDBTable, strlwr(p), &dummy);
        Tcl_SetHashValue(hPtr, ePtr);
	free(p);
    }

}
/*
 * GetColorByName
 *
 */

Bool
GetColorByName(colorname, color)
char *colorname;
XColor *color;
{
    Tcl_HashEntry *hPtr;
    ColorEntry *cPtr;
    HDC hdc;
    char *p;

    if (!builded) {
        BuildColorDB();
    }
    p = malloc(strlen(colorname) + 1);
    strcpy(p, colorname);
    hPtr = Tcl_FindHashEntry(&ceDBTable, (char *) strlwr(p));
    free(p);

    if (hPtr == NULL)
        return 0;	 

    cPtr = (ColorEntry *) Tcl_GetHashValue(hPtr);
    hdc = GetDC((HWND)NULL);
    color->pixel = RGB(cPtr->red, cPtr->green, cPtr->blue);
    if (pure) {
	color->pixel = GetNearestColor(hdc, color->pixel);
    }
    ReleaseDC((HWND)NULL, hdc);
    color->red = GetRValue(color->pixel);
    color->green = GetGValue(color->pixel);
    color->blue = GetBValue(color->pixel);
    color->pad = 0;

    return 1;
}      

/*
 * GetColorByValue
 *
 */

Bool
GetColorByValue(value, color)
char *value;
XColor *color;
{
	char fmt[16];
	int i;
	HDC hdc;

	if ((i = strlen(value+1))%3)
		return 0;
	i /= 3;

	sprintf(fmt, "%%%dx%%%dx%%%dx", i, i, i);
	sscanf(value+1, fmt, &color->red, &color->green, &color->blue);
	color->pad = 0;
    hdc = GetDC((HWND)NULL);
    color->pixel = RGB(color->red, color->green, color->blue);
    if (pure) {
	color->pixel = GetNearestColor(hdc, color->pixel);
    }
    color->red = GetRValue(color->pixel);
    color->green = GetGValue(color->pixel);
    color->blue = GetBValue(color->pixel);
      ReleaseDC((HWND)NULL, hdc);

	return 1;
}

Bool
GetColorByXColor(color)
XColor *color;
{
    HDC hdc;

    hdc = GetDC((HWND)NULL);
    color->pixel = RGB(color->red, color->green, color->blue);
    if (pure) {
        color->pixel = GetNearestColor(hdc, color->pixel);
    }
    ReleaseDC((HWND)NULL, hdc);

	return 1;
}
/*
 * Tb_SetClipboard --
 *
 */

void
__far __pascal
Twin_SetClipboard(Tk_Window tkwin)
{
    HGLOBAL hGlobal;
    char __far * pGlobal;
    char __far * pString;
    HWND hwnd = Tk_WindowId(tkwin);
    GetInfo getInfo;
    Tcl_Interp *interp;
    int result;

    interp = Tcl_CreateInterp();
    getInfo.string = (char *) ckalloc(100);
    getInfo.bytesAvl = 100;
    getInfo.bytesUsed = 0;
    result = Tk_GetSelection(interp, tkwin, XA_STRING, SelGetProc,
	    (ClientData) &getInfo);
    if (result != TCL_OK) {
        ckfree(getInfo.string);
	return;
    }
   Tcl_DeleteInterp(interp);
    if ((hGlobal = GlobalAlloc(GHND, strlen(getInfo.string) + 1)) == None) {
    	return;
    }
    pGlobal = GlobalLock(hGlobal);
    pString = getInfo.string;

    while(*pGlobal++ = *pString++) ;
    GlobalUnlock(hGlobal); 
    if (GetClipboardOwner() == hwnd) { 
	OpenClipboard(hwnd);
	SendMessage(hwnd, WM_USER+1, 0, 0L);
	EmptyClipboard();
  	SetClipboardData(CF_TEXT, hGlobal);
	CloseClipboard();
    }
    ckfree(getInfo.string);
}

int
__far __pascal
Twin_GetClipboardData(interp, proc, clientData)
    Tcl_Interp *interp;		/* Interpreter to use for reporting
				 * errors. */
    Tk_GetSelProc *proc;	/* Procedure to call to process the
				 * selection, once it has been retrieved. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    HGLOBAL hGlobal;
    int result;
    char *pGlobal;

    if (OpenClipboard((HWND) NULL)) {
	if ((hGlobal = GetClipboardData(CF_TEXT)) == (HGLOBAL) NULL) {
	    CloseClipboard();
	return TCL_ERROR;
	}
	pGlobal = GlobalLock(hGlobal);
	result = (*proc)(clientData, interp, pGlobal);
	GlobalUnlock(hGlobal);
	CloseClipboard();
	return result;
    }
    return TCL_ERROR;
}
/*
 * reberth
 */

static void
reberth(bits)
unsigned char *bits;
{
    unsigned char buffer[4];
    int i;

    for (i = 0; i < 64; i+= 4) {
    	memcpy(buffer, bits + i, 4);
	memcpy(bits + i, bits + (124 - i), 4);
	memcpy(bits + (124 - i), buffer, 4);
    }
} 

/*
 *	LoadCursorRuntime --
 *
 */
int 
Twin_ReadCursor(filename, width, height, xhot, yhot)
char *filename;
unsigned int *width, *height;
int *xhot, *yhot;
{
	struct _stat st;
	char *buffer;
	FILE *fp;
	int *src, *mask;
	CURSORFILEHEADER *csfh;
	HCURSOR hcs;

	if (_stat(filename, &st) != 0)	
		return None;  
	if ((fp = fopen(filename, "rb")) == NULL)
		return None;
	if ((buffer = (char *)malloc((size_t)st.st_size)) == NULL)
		return None;
	if (fread(buffer, (size_t)st.st_size, 1, fp) != 1)	
		return None;
	fclose(fp);
	csfh = (CURSORFILEHEADER *)buffer;
	*width = 32;
	*height = 32;
	if (!xhot) *xhot = csfh->x_hot;
	if (!yhot) *yhot = csfh->y_hot;

        src = (int *)(buffer + HDR_SIZE);					
	reberth(src);
	mask = (int *)(buffer + HDR_SIZE + BM_SIZE);
	reberth(mask);
	hcs = CreateCursor(GetAppInst(), csfh->x_hot, csfh->y_hot, 
			32, 32, mask, src);	

	if (hcs == None)
		return None;
			
	free(buffer);

	return hcs;
}	

HCURSOR 
GetCursorByShape(const char *shape)
{
    return LoadCursor(0, shape);
}

#ifdef KANJI 
int FAR PASCAL EXPORT
Twin_WSTextWidth(font, wstr, gmask, len)
     XFontStruct *font;
     wchar *wstr;
     int gmask;
     int len;
{
#define bufsize	256
    wchar       *cp;
    wchar       *wstr1 = wstr + len;
    int         width = 0;

    while( wstr < wstr1 ) {
	cp = wstr;

	if( font == NULL ) {
	    while( wstr < wstr1) wstr++;
	    continue;
	}

	while( wstr < wstr1) {
	    wstr++;
	}

	if( cp == wstr ) continue;

	/* flush */
	width += XTextWidth16(font, (XChar2b *) cp, wstr - cp);
    }

    return width;
}
#endif /* TB */

void FAR PASCAL
Twin_Scroll(hwnd, xSrc, ySrc,	cxSrc, cySrc, xDist, yDist)
HWND hwnd;
int xSrc, ySrc, xDist, yDist;
UINT cxSrc, cySrc;
{
//    HDC hdc;
    RECT rc;

    rc.top = ySrc;
    rc.left = xSrc;
    rc.right = xSrc + cxSrc;
    rc.bottom  = ySrc + cySrc;
//    hdc = GetDC(hwnd);
//    ScrollDC(hdc, - (xSrc - xDist), - (ySrc - yDist), &rc, NULL, NULL, NULL);
    ScrollWindow(hwnd, - (xSrc - xDist), - (ySrc - yDist), &rc, NULL);
//    ReleaseDC(hwnd, hdc);
}

HICON FAR PASCAL
Twin_LoadIcon(LPSTR lpszIconName)
{
    LPCSTR pszIcon;
    int len = _fstrlen(lpszIconName);

    if (!_fstrcmp(lpszIconName, "IDI_APPLICATION"))
	pszIcon = IDI_APPLICATION;
    else if (!_fstrcmp(lpszIconName, "IDI_ASTERISK"))
	pszIcon = IDI_ASTERISK;
    else if (!_fstrcmp(lpszIconName, "IDI_EXCLAMATION"))
	pszIcon = IDI_EXCLAMATION;
    else if (!_fstrcmp(lpszIconName, "IDI_HAND"))
	pszIcon = IDI_HAND;
    else if (!_fstrcmp(lpszIconName, "IDI_QUESTION"))
	pszIcon = IDI_QUESTION;
    else 
	return 0;
    return LoadIcon(0, pszIcon);
}
/*
 *----------------------------------------------------------------------
 *
 * Twin_GetLong --
 *
 *	Given a string,  produce  the corresponding  lonng integer 
 *      value.
 *
 * Results:
 *	The return value is normally TCL_OK;  in this case *intPtr
 *	will be set to the integer value equivalent to string.  If
 *	string is improperly formed then TCL_ERROR is returned and
 *	an error message will be left in interp->result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
FAR PASCAL
Twin_GetLong(interp, string, longPtr)
    Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
    char *string;		/* String containing a (possibly signed)
				 * integer in a form acceptable to strtol. */
    long *longPtr;		/* Place to store converted result. */
{
    char *end, *p;
    long l;

    /*
     * Note: use strtoul instead of strtol for integer conversions
     * to allow full-size unsigned numbers, but don't depend on strtoul
     * to handle sign characters;  it won't in some implementations.
     */

    for (p = string; isspace(UCHAR(*p)); p++) {
	/* Empty loop body. */
    }
    if (*p == '-') {
	l = - (long) strtoul(p+1, &end, 0);
    } else if (*p == '+') {
	l = (long) strtoul(p+1, &end, 0);
    } else {
	l = (long) strtoul(p, &end, 0);
    }
    while ((*end != '\0') && isspace(UCHAR(*end))) {
	end++;
    }
    if ((end == string) || (*end != 0)) {
	Tcl_AppendResult(interp, "expected integer but got \"", string,
		"\"", (char *) NULL);
	return TCL_ERROR;
    }
    *longPtr = l;
    return TCL_OK;
}
 