/*
 * streem.c -- 
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include "twinint.h"
#include <stdlib.h>
#include <string.h>

struct tagStreem {
    char *string;
    int length;
    struct tagStreem *next;
};

typedef struct tagStreem Streem;

Streem *streemHead = NULL;
Streem *streemTail = NULL;

void FAR PASCAL
FillStreem(const char *string)
{
    Streem *sPtr;
    
    sPtr = (Streem *) malloc(sizeof(Streem));
    sPtr->length = strlen(string);
    sPtr->string = (char *)malloc(sPtr->length + 1);
    sPtr->next = NULL;
    strcpy(sPtr->string, string);
    if (streemHead == NULL) {
        streemHead = sPtr;
    } else {
        streemTail->next = sPtr;
    }
    streemTail = sPtr;
}

int FAR PASCAL
ReadStreem(char *buffer, int length)
{
    Streem *sPtr;
    int rest = length;
    char *bPtr = buffer;
    int cnt = 0;
    
    for(sPtr = streemHead; ; sPtr = sPtr->next) {
        if (rest == 0) break;
        if (sPtr == NULL) {
	    if (cnt > 0) return cnt;
            WaitInput();
	    if (streemHead  == NULL) return 0;
	    sPtr = streemHead;
        }
        if (sPtr->length > rest) {
	    char *new;

            strncpy(bPtr, sPtr->string, rest);
	    cnt += rest;
            sPtr->length = sPtr->length - rest;
            new = (char *) malloc(sPtr->length + 1);
            strcpy(new, sPtr->string + length);
            free(sPtr->string);
            sPtr->string = new;
	    break;
        } else {
            strncpy(bPtr, sPtr->string, sPtr->length);
	    cnt += sPtr->length;
            rest = rest - sPtr->length;
            bPtr = bPtr + sPtr->length;
            streemHead = sPtr->next;
            free(sPtr->string);
            free(sPtr);
        }
    }
    if (sPtr == NULL) {
        streemHead = streemTail = NULL;
    } else {
        streemHead = sPtr;
    }
    return cnt;
}

int FAR PASCAL
FlushStreem(void)
{
    Streem *sPtr;
    
    for (sPtr = streemHead; sPtr != NULL; sPtr = streemHead) {
        streemHead = sPtr->next;
        free(sPtr->string);
        free(sPtr);
    }
    streemTail = NULL;
    return 0;
}
  
void NEAR PASCAL
WaitInput(void)
{
    MSG msg;
    
    while(GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
	if (msg.message == WM_USER) break;
    }
}

int FAR PASCAL
WriteStreem(char * sp, int n)
{
    int i = n;
    while(i--) {
	SendMessage(GetHomeWnd(), WM_USER+20, (WPARAM)*sp++ , 0L);
    }
    return n;
}

void FAR PASCAL
wputs(const char *s, HWND hwnd)
{
    for(; *s != '\0'; s++) {
    	if  (*s	== '\n') {
	    SendMessage(hwnd, WM_USER+20, (WPARAM)'\r', 0L);
	} else {
	    SendMessage(hwnd, WM_USER+20, (WPARAM)*s, 0L);
	}
    }
}
