/*
 * wndproc.c --
 *
 * Copyright (c) 1995 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include "twinint.h"
#include <x11/xutil.h>
#include <string.h>

/* 
 * XSetWindowAttributes
 */

XSetWindowAttributes defAttrs = {
    None,		/* backfround pixmap */ 
    RGB(255, 255, 255),	/* background pixel */
    None, 		/* border pixmap */	
    0L, 		/* border pixel */
    ForgetGravity, 	/* bit gravity */
    NorthWestGravity, 	/* window gravity */
    NotUseful,		/* backing store */
    0, 			/* backing planes */
    0, 			/* backing pixel */
    False, 		/* save under */
    0L, 		/* even tmask */
    0L, 		/* do not propagate mask */
    False, 		/* override redirect */
    None, 		/* colormap */
    None		/* cursor */
};

/* 
 * XWindowChanges
 */

XWindowChanges defChanges = {
    0, 		/* x */
    0, 		/* y */
    1, 		/* width */
    1, 		/* height */
    1, 		/* border width */
    None, 	/* sibling */
    Above	/* stack mode */
};
int initiarized = 0;
int cxMinTrackSize;
int cyMinTrackSize;
int cIdTopLevel = -1;


long FAR PASCAL DefTwinProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

/****************************************************************************** 
 * RegisterTopLevelClass -- 
 *****************************************************************************/

void FAR PASCAL 
RegisterTopLevelClass(HINSTANCE hInstance)
{
    WNDCLASS wndclass;
    char szClassName[16];

    sprintf(szClassName, "TopLevel%d", ++cIdTopLevel);
    wndclass.style	   = cIdTopLevel ? CS_SAVEBITS : CS_SAVEBITS|CS_GLOBALCLASS;
    // |CS_BYTEALIGNWINDOW; CS_HREDRAW | CS_VREDRAW ;
    wndclass.lpfnWndProc   = TopLevelProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 32;
    wndclass.hInstance     = hInstance;
    wndclass.hIcon	   = LoadIcon(NULL, IDI_APPLICATION);
    wndclass.hCursor       = NULL;
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = szClassName;

    RegisterClass(&wndclass);
}

/****************************************************************************** 
 * TopLevelProc --
 *****************************************************************************/
long FAR PASCAL
TopLevelProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    HDC hdc;
    HGLOBAL hXSWAttrs, hWork;
    XSetWindowAttributes * pXSWAttrs;
    HBRUSH hBrush;
    RECT rc;

    switch(message) {
    case WM_GETMINMAXINFO:
    	if((hWork = GetProp(hwnd, "XA_WM_NORMAL_HINTS")) != NULL) {
	    XSizeHints * pHints;
	    MINMAXINFO FAR* lpmmi;

	    lpmmi = (MINMAXINFO FAR*) lParam;
	    if (!initiarized) {
	        initiarized = 1;
	    	cxMinTrackSize = lpmmi->ptMinTrackSize.x;
		cyMinTrackSize = lpmmi->ptMinTrackSize.y;
	    }
	    pHints = (XSizeHints *) GlobalLock(hWork);

    	    hXSWAttrs = GetWindowWord(hwnd, XSWA_OFFSET);
	    pXSWAttrs = (XSetWindowAttributes *) GlobalLock(hXSWAttrs);
	    if (!(pHints->flags & USSize)) {
	    	if (pHints->flags & PMinSize) {
	    	    lpmmi->ptMinTrackSize.x =
	    	    	    AdjustWidth(pHints->min_width, pXSWAttrs->override_redirect);
		    lpmmi->ptMinTrackSize.y = 
			    AdjustHeight(pHints->min_height, pXSWAttrs->override_redirect);
	    	}
	    	if (pHints->flags & PMaxSize) {
	    	    lpmmi->ptMaxTrackSize.x = 
			    AdjustWidth(pHints->max_width, pXSWAttrs->override_redirect);
		    lpmmi->ptMaxTrackSize.y = 
			    AdjustHeight(pHints->max_height, pXSWAttrs->override_redirect);
	    	}
	    }
	    GlobalUnlock(hWork);
	    GlobalUnlock(hXSWAttrs);
	}
	break;

    case WM_SIZE:
    	switch (wParam) {
    	case SIZE_MINIMIZED:
	    hWork = GetProp(hwnd, "XA_WM_ICON_NAME"); break;
	case SIZE_MAXIMIZED:
	case SIZE_RESTORED: 
	    hWork = GetProp(hwnd, "XA_WM_NAME"); break;
	default:
	    hWork = NULL;
	}
	if (hWork != NULL) {
	    SetWindowText(hwnd, (LPSTR) GlobalLock(hWork));
    	    GlobalUnlock(hWork);
	}
	return 0;

    case WM_NCLBUTTONUP:	case WM_NCMBUTTONUP:	    case WM_NCRBUTTONUP:
    case WM_NCLBUTTONDOWN:	case WM_NCMBUTTONDOWN:	    case WM_NCRBUTTONDOWN:
    case WM_NCLBUTTONDBLCLK:	case WM_NCMBUTTONDBLCLK:    case WM_NCRBUTTONDBLCLK:
       	if (fPtGrab) 
	    return 0;
	break;

    case WM_CLOSE:
    	PostAppMessage(GetCurrentTask(), CLIENTMESSAGE, hwnd, 0L);
	return 0;

    case WM_PAINT:
	hdc = GetDC(hwnd); 
	hXSWAttrs = GetWindowWord(hwnd, XSWA_OFFSET);
    	pXSWAttrs = (XSetWindowAttributes *) GlobalLock(hXSWAttrs);
	hBrush = CreateSolidBrush(pXSWAttrs->background_pixel);
	GlobalUnlock(hXSWAttrs);
	GetUpdateRect(hwnd, &rc, FALSE);
	FillRect(hdc, &rc, hBrush);
	DeleteObject(hBrush);
	ReleaseDC(hwnd, hdc);
	return 0;
    }
    return DefTwinProc(hwnd, message, wParam, lParam);
}

/****************************************************************************** 
 * RegisterWidgetClass -- 
 *****************************************************************************/
void FAR PASCAL
RegisterWidgetClass(HINSTANCE hInstance)
{
    WNDCLASS wndclass;

    wndclass.style	   = CS_DBLCLKS|CS_GLOBALCLASS|CS_HREDRAW|CS_VREDRAW;
    //|CS_BYTEALIGNWINDOW;
    wndclass.lpfnWndProc   = WidgetProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 32;
    wndclass.hInstance     = hInstance;
    wndclass.hIcon	   = LoadIcon(0, IDI_APPLICATION);
    wndclass.hCursor       = NULL;
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = "Widget";

    RegisterClass(&wndclass);
}

/****************************************************************************** 
 * WidgetProc --
 *****************************************************************************/
long FAR PASCAL
WidgetProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    HGLOBAL hXSWAttrs;
    XSetWindowAttributes * pXSWAttrs;
    RECT rc, rc2;
    HBRUSH hbr;
    HDC hdc;
    static int flag = 0;

    switch(message) {
    case WM_USER+1:
        flag = 1;
	return 0;

    case WM_DESTROYCLIPBOARD:
        if (flag) {
	    flag = 0;
	    return 0;
	}
	PostMessage(hwnd, SELECTIONCLEAR, wParam, lParam);
	return 0;

    case WM_PAINT:
	GetUpdateRect(hwnd, &rc, FALSE);
	GetWindowRect(hwnd, &rc2);

	if ((rc.top != 0 && rc.left != 0) 
		|| rc.bottom < (rc2.bottom - rc2.top) 
		|| rc.right < (rc2.right - rc2.left) )  {
	    hXSWAttrs = GetWindowWord(hwnd, XSWA_OFFSET);
    	    pXSWAttrs = (XSetWindowAttributes *) GlobalLock(hXSWAttrs);
	    hbr = CreateSolidBrush(pXSWAttrs->background_pixel);
	    hdc = GetDC(hwnd); 
	    FillRect(hdc, &rc, hbr);
	    ReleaseDC(hwnd, hdc);
	    DeleteObject(hbr);
	    GlobalUnlock(hXSWAttrs);
	}
	return 0;
        	 
    }
    return DefTwinProc(hwnd, message, wParam, lParam);
}

/****************************************************************************** 
 * DefTwinProc -- 
 *****************************************************************************/
long FAR PASCAL
DefTwinProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    HGLOBAL hXSWAttrs, hXWChanges, hWork;
    XSetWindowAttributes * pXSWAttrs;
    XWindowChanges * pXWChanges;
    WINDOWPOS FAR * lpwp;
    HCURSOR hcs;
    static int flag = 0;

    switch(message) {
    case WM_COMMAND:
    	switch (wParam) {
	case IDM_CURSORCHANGED:
	    if (GetWindowWord(hwnd, HCCS_OFFSET) == NULL) 
	    	SetWindowWord(hwnd, HACS_OFFSET, HIWORD(lParam));
	    break;
	case IDM_CHANGEICON:
	    break;
	}
	return 0;
	    
    case WM_CREATE:
    	hXSWAttrs = GlobalAlloc(GHND, sizeof(XSetWindowAttributes));
	pXSWAttrs = (XSetWindowAttributes *) GlobalLock(hXSWAttrs);
	*pXSWAttrs = defAttrs;
	GlobalUnlock(hXSWAttrs);
	hXWChanges = GlobalAlloc(GHND, sizeof(XWindowChanges));
	pXWChanges = (XWindowChanges *) GlobalLock(hXWChanges);
	*pXWChanges = defChanges;
	GlobalUnlock(hXWChanges);

    	SetWindowWord(hwnd, XSWA_OFFSET, (WORD) hXSWAttrs);
	SetWindowWord(hwnd, XWCH_OFFSET, (WORD) hXWChanges);
	PostMessage(hwnd, CREATENOTIFY,  wParam, lParam); 
	return 0;

    case WM_SETFOCUS:
	PostMessage(hwnd, FOCUSIN, wParam, lParam); 
	return 0;
	
    case WM_KILLFOCUS:
	PostMessage(hwnd, FOCUSOUT, wParam, lParam); 
	return 0;

    case WM_MOUSEACTIVATE:
       	if (fPtGrab) {
	    return MA_NOACTIVATE;
	}
	break;
	 
    case WM_MOUSEMOVE: 
    	if ((hcs = GetWindowWord(hwnd, HACS_OFFSET)) == NULL) {
            if ((hcs = GetWindowWord(hwnd, HCCS_OFFSET)) == NULL)
	    	hcs = FindCursor(hwnd);
	    SetWindowWord(hwnd, HACS_OFFSET, hcs);
	}
	SetCursor(hcs);				    
	return 0; 

    case WM_DESTROY: 
    	GlobalFree(GetWindowWord(hwnd, XSWA_OFFSET));
	GlobalFree(GetWindowWord(hwnd, XWCH_OFFSET));
	EnumProps(hwnd, ClearProps);
	return 0;

    case WM_WINDOWPOSCHANGED:	 

    	hWork = GlobalAlloc(LHND, sizeof(WINDOWPOS));
	lpwp = (WINDOWPOS *) GlobalLock(hWork);
	memcpy(lpwp, (WINDOWPOS *)lParam, sizeof(WINDOWPOS));

	hXWChanges = GetWindowWord(hwnd, XWCH_OFFSET);
	pXWChanges = (XWindowChanges *) GlobalLock(hXWChanges);

    	hXSWAttrs = GetWindowWord(hwnd, XSWA_OFFSET);
	pXSWAttrs = (XSetWindowAttributes *) GlobalLock(hXSWAttrs);
	if (lpwp->flags & SWP_NOMOVE) {
	     lpwp->x = pXWChanges->x;
	     lpwp->y = pXWChanges->y;
	} else {
	     pXWChanges->x = lpwp->x = ReAdjustX(lpwp->x, pXSWAttrs->override_redirect);
	     pXWChanges->y = lpwp->y = ReAdjustY(lpwp->y, pXSWAttrs->override_redirect);
	}
	if (lpwp->flags & SWP_NOSIZE) {
	     lpwp->cx = pXWChanges->width;
	     lpwp->cy = pXWChanges->height;
	} else {

	     pXWChanges->width = lpwp->cx = ReAdjustWidth(lpwp->cx, pXSWAttrs->override_redirect);
	     pXWChanges->height = lpwp->cy = ReAdjustHeight(lpwp->cy, pXSWAttrs->override_redirect);
	}
	GlobalUnlock(hWork);
	GlobalUnlock(hXWChanges);
	GlobalUnlock(hXSWAttrs);

	PostMessage(hwnd, CONFIGURENOTIFY, hWork, 0L); 

	break;

#if 0
    case WM_ACTIVATE:	 /* top */
	if (wParam != WA_INACTIVE) { 
   	    PostMessage(hwnd, EXPOSE, 0, 0L);
 	    EnumChildWindows(hwnd, MakeProcInstance(SendActivate, GetAppInst()), 0L);
	}
	break;		/* widgit */
    case WM_ACTIVATE:
    	PostMessage(hwnd, EXPOSE, 0, 0L);
	return 0;
#endif
    }
    return DefWindowProc(hwnd, message, wParam, lParam);
}