/*
 * xdisplay.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include "twinint.h" 
#include <X11/Xlib.h>
#include <io.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

static Display FAR * lpDispHead = NULL;

static Screen FAR * NEAR PASCAL mkScreen(Display *);

/*
 *	XOpenDisplay --
 *
 */

Display * FAR PASCAL
XOpenDisplay(LPCSTR lpszScreenName)
{
    Display FAR * lpdisp;
    Screen FAR * lpscr;

    if (lpDispHead != NULL) {
	return lpDispHead;
    }
    
    lpdisp = (Display *) malloc(sizeof(Display));
    if (lpdisp == NULL) {
	return NULL;
    }
    lpscr = mkScreen(lpdisp);
    if (lpscr == NULL) {
    	free(lpdisp);
        return NULL;
    }

    lpdisp->fd = GetCurrentTask();
    lpdisp->head = NULL;
    lpdisp->tail = NULL;		
    lpdisp->qlen = 0;   

    if (lpszScreenName != NULL && *lpszScreenName != '\0') {
    	lpdisp->display_name = (char *)malloc(strlen(lpszScreenName) + 1);
	if (lpdisp->display_name == NULL) {    
	    free(lpscr);
	    free(lpdisp);
	    return NULL;
	}
	strcpy(lpdisp->display_name, lpszScreenName);		
    } else {
    	lpdisp->display_name = "";
    }
    lpdisp->default_screen = 0;     	/* default screen for operations */
    lpdisp->nscreens = 1;           	/* number of screens on this server*/
    lpdisp->screens = lpscr;        	/* pointer to list of screens */

    lpdisp->next = lpDispHead;
    lpDispHead = lpdisp;
    return lpdisp;
}

static Screen *	NEAR PASCAL
mkScreen(Display *display)
{
    Screen *lpscr;
    HDC hdc;

    if ((lpscr = malloc(sizeof(Screen))) == NULL)
	return NULL;

    lpscr->display = display;				/* back pointer to display structure */
    lpscr->root = HWND_DESKTOP;            		/* Root window id. */
    lpscr->width = GetSystemMetrics(SM_CXSCREEN);	/* width and height of screen */
    lpscr->height = GetSystemMetrics(SM_CYSCREEN);      

    hdc = GetDC(HWND_DESKTOP);
    lpscr->root_depth = GetDeviceCaps(hdc, BITSPIXEL);        
    lpscr->mwidth = GetDeviceCaps(hdc, HORZSIZE); 	/* width and height of  in millimeters */
    lpscr->mheight = GetDeviceCaps(hdc, VERTSIZE);    
    ReleaseDC(HWND_DESKTOP, hdc);	  
    lpscr->cmap = 1;          				/* default color map */
    lpscr->white_pixel = RGB(255, 255, 255);		/* White and Black pixel values */
    lpscr->black_pixel = RGB(0, 0, 0);			
    lpscr->backing_store = False;			/* Never, WhenMapped, Always */
    lpscr->save_unders = False;

    return lpscr;
}


/*
 *	XCloseDisplay --
 *
 *	
 */

void  FAR PASCAL
XCloseDisplay(Display *display)
{
    if (lpDispHead != NULL && lpDispHead == display) {
        free(display->screens);
	free(display);
	lpDispHead = NULL;
    }
}

/*
 * 	XRootWindow, XFlush, XSync
 */

Window FAR PASCAL
XRootWindow(Display *display, int screen_number)
{
    if (lpDispHead != NULL && lpDispHead == display 
    	    && screen_number <= display->nscreens) {
	return (display->screens[screen_number]).root;
    }
    return None;
}

void FAR PASCAL
XFlush(Display *dummy)
{
}

void FAR PASCAL
XSync(Display *dummy1, Bool dummy2)
{
}

void FAR PASCAL
XGrabServer(Display FAR * dummy)
{
}

void FAR PASCAL
XUngrabServer(Display FAR * dummy)
{
}

XHostAddress FAR * FAR PASCAL
XListHosts(Display * dummy, int FAR * lpnHosts, Bool FAR * lpenable)
{
    *lpnHosts = 0;
    *lpenable = True;
    return NULL;
}
