/*
 * xfont.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include <twinint.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>
#ifdef KANJI 
#include <mbstring.h>
#endif /* KANJI */

#define wchar short

#if 1

static LOGFONT defLogFont = {
    14,				/* lfHeight */
    0,				/* lfWidth */
    0, 				/* lfEscapement */
    0,				/* lfOrientation */
    FW_DONTCARE,		/* lfWeight */
    0, 				/* lfItalic */
    0,				/* lfUnderline */
    0,				/* lfStrikeOut */
    DEFAULT_CHARSET,		/* lfCharSet */
    OUT_TT_PRECIS,		/* lfOutPrecision */
    CLIP_DEFAULT_PRECIS,	/* lfClipPrecision */
    PROOF_QUALITY,		/* lfQuality */
    DEFAULT_PITCH|FF_DONTCARE,	/* lfPitchAndFamily */
    {'\0'},			/* lfFaceName[LF_FACESIZE] */
};

typedef struct {
char fncopy[128];
int nfield;
LPCSTR field[14];
} FONTENT;

int
GetFontEnt (LPCSTR lpszFontName, FONTENT FAR * lpFontEnt)
{
    register LPSTR p = lpFontEnt->fncopy;
    int i = 0;

    _fstrcpy(lpFontEnt->fncopy, lpszFontName);
    lpFontEnt->field[i++] = (*p == '-') ? ++p : p;
    p = strchr(p, '-');	
    for (; p != NULL; i++, p = strchr(p, '-')) {
      if (*p == '-') *p = '\0';
      if (++p == '\0') 
          break;
      lpFontEnt->field[i] = p;
    }	
    if (*(lpszFontName + strlen(lpszFontName) - 1) == '*') {
        while (i < 14) 
	     lpFontEnt->field[i++] = "*";
    }
    lpFontEnt->nfield = i;
    return i;
}

static LOGFONT * NEAR PASCAL
NameToLogFont(const char *name, LOGFONT * lplf)
{
    FONTENT fe;
    int nfield;

    *lplf = defLogFont;
    lplf->lfCharSet= DEFAULT_CHARSET;
    if ((nfield = GetFontEnt(name, &fe)) == -1) {
	return NULL;
    } else if (nfield == 1) { /* specify fot directry */
    } else if (nfield == 3) { /* twin style */
	_fstrcpy(lplf->lfFaceName, (LPSTR)fe.field[0]);
#ifdef KANJI
	if (_ismbblead((unsigned char) *fe.field[0]))
	        lplf->lfCharSet= SHIFTJIS_CHARSET;
#endif
        lplf->lfWeight = strstr(fe.field[1], "Bold")? FW_BOLD: FW_MEDIUM;
        lplf->lfItalic = strstr(fe.field[1], "Italic") != NULL;
        lplf->lfHeight = - atoi(fe.field[2]); 
    } else if (nfield == 14) { /* tkwin style */
	strcpy(lplf->lfFaceName, "Times New Roman");
        if (!_fstrcmp(fe.field[2], "bold") || !_fstrcmp(fe.field[2], "Bold"))
	    lplf->lfWeight = FW_BOLD;
        else if (!_fstrcmp(fe.field[2], "medium") || !_fstrcmp(fe.field[2], "Medium")) 
	    lplf->lfWeight = FW_MEDIUM;
        else if (!_fstrcmp(fe.field[2], "light") || !_fstrcmp(fe.field[2], "Light")) 
	    lplf->lfWeight = FW_LIGHT;
        else if (!_fstrcmp(fe.field[2], "thin") || !_fstrcmp(fe.field[2], "Thin")) 
	    lplf->lfWeight = FW_THIN;
        else
	    lplf->lfWeight = FW_NORMAL;
        switch (*fe.field[3]) {
        case 'o': case 'O':
        case 'i': case 'I':
            lplf->lfItalic = 1;
	    break;
        case 'r': case 'R':
	case '*':
	default:
            lplf->lfItalic = 0;
	    break;	
	}
        if (*fe.field[5] == '\0' || *fe.field[5] == '*') {
	    lplf->lfPitchAndFamily |= FF_DONTCARE;    
        } else {  
      	    if (!_fstrcmp(fe.field[5], "sans") || !_fstrcmp(fe.field[5], "Sans"))
	    	lplf->lfPitchAndFamily |= FF_SWISS;
            else if (!_fstrcmp(fe.field[5], "serif") || !_fstrcmp(fe.field[5], "Serif")) 
	    	lplf->lfPitchAndFamily |= FF_ROMAN;
            else 
	    	lplf->lfPitchAndFamily |= FF_DONTCARE;
	}
        lplf->lfHeight = - atoi(fe.field[6]);
        {
            HDC hdc = GetDC(HWND_DESKTOP); 
            int ppi = GetDeviceCaps(hdc, LOGPIXELSY);
            lplf->lfHeight = - (int) ((atoi(fe.field[7]) / 10.0) * ( ppi / 72.0));
	    ReleaseDC(HWND_DESKTOP, hdc);
	}
        switch (*fe.field[10]) {
	case 'm': case 'M':
	    lplf->lfPitchAndFamily |= FIXED_PITCH;
	    break;
	case 'p': case 'P':
	    lplf->lfPitchAndFamily |= VARIABLE_PITCH;
	    break;
	default:
	    lplf->lfPitchAndFamily |= DEFAULT_PITCH;
	}
        lplf->lfWidth = - atoi(fe.field[11]);
    }
    return lplf;
}
#else
static LOGFONT * NEAR PASCAL
NameToLogFont(const char *name, LOGFONT * lplf)
{
    const char *pHead, *pTail;
#ifdef KANJI
    const char *lp;
#endif /* KANJI */

    memset(lplf, '\0', sizeof(LOGFONT));

    pHead = name;
#ifdef KANJI
    if ((pTail = (char *) _mbschr((const unsigned char *) pHead, '-')) == NULL) 
    	goto default_font;
    strncpy(lplf->lfFaceName, pHead, pTail - pHead);

    lplf->lfCharSet= ANSI_CHARSET;
    for (lp = pHead; lp != pTail; lp++) {
    	if (_ismbblead((unsigned char) *lp)) {
    	    lplf->lfCharSet= SHIFTJIS_CHARSET;
	    break;
	}
    }

    pHead = pTail + 1;

    if ((pTail = (char *) _mbschr((const unsigned char *) pHead, '-')) == NULL) 
    	goto default_font;
#else
    if ((pTail = strchr((const unsigned char *) pHead, '-')) == NULL) 
    	goto default_font;
    strncpy(lplf->lfFaceName, pHead, pTail - pHead);

    lplf->lfCharSet= ANSI_CHARSET;

    pHead = pTail + 1;

    if ((pTail = strchr((const unsigned char *) pHead, '-')) == NULL) 
    	goto default_font;
#endif
    lplf->lfWeight = strstr(pHead, "Bold")? FW_BOLD: FW_MEDIUM;
    lplf->lfItalic = strstr(pHead, "Italic") != NULL;
    pHead = pTail + 1;    

    if ((lplf->lfHeight = - atoi(pHead)) ==0) 
    	goto default_font;

    return lplf;

default_font:    
    lplf->lfHeight = 14;
    return lplf;
}
#endif /* 0 */

XFontStruct * FAR PASCAL
XLoadQueryFont(Display *dummy, const char *name)
{
	XFontStruct *fs = (XFontStruct *)malloc(sizeof(XFontStruct));
	LOGFONT lf;
	HDC hdc;
	HFONT hfont_old;
	TEXTMETRIC tm;
	XCharStruct cs;
	memset(fs, 0, sizeof(XFontStruct)); 
	memset(&lf, 0, sizeof(LOGFONT)); 
	fs->fid = CreateFontIndirect(NameToLogFont(name, &lf) );
	hdc = CreateCompatibleDC((HDC)NULL);
	hfont_old = SelectObject( hdc, fs->fid );

	GetTextMetrics( hdc, &tm );
/*
	fs->direction = (lf.lfEscapement == 900? FontRightToLeft: FontLeftToRight);
*/
	fs->min_byte1 = fs->max_byte1 = 0;
	fs->min_char_or_byte2 = tm.tmFirstChar;
	fs->max_char_or_byte2 = tm.tmLastChar;
	fs->all_chars_exist = TRUE;
	fs->default_char = tm.tmDefaultChar;

	    fs->ascent = tm.tmAscent;
	    fs->descent = tm.tmDescent;
	cs.width = tm.tmAveCharWidth;
	cs.ascent = fs->ascent;
	cs.descent = fs->descent;
	cs.lbearing = 0;
	cs.rbearing = cs.width;
	fs->min_bounds = fs->max_bounds = cs;
	fs->per_char = NULL;

	if( tm.tmAveCharWidth != tm.tmMaxCharWidth ) {
		int i, nchars = tm.tmLastChar - tm.tmFirstChar + 1;
		int *wbuf = (int *)malloc(sizeof(int) * nchars);
		int min_width = 30000;
#if 1
		GetCharWidth( hdc, tm.tmFirstChar, tm.tmLastChar, (int FAR *)wbuf );

		fs->per_char = (XCharStruct *)malloc(sizeof(XCharStruct) * nchars);
		for( i = 0; i < nchars ; i++ ) {
			fs->per_char[i] = cs;
			fs->per_char[i].width = wbuf[i];
			if( min_width > wbuf[i] )
				min_width = wbuf[i];
		}
#endif
		fs->max_bounds.width = tm.tmMaxCharWidth;
		fs->min_bounds.width = min_width;
	}

	SelectObject( hdc, hfont_old );
	DeleteDC( hdc );

	return fs;
}

void FAR PASCAL
XFreeFont(Display *dummy, XFontStruct *font_struct)
{
	if( font_struct->per_char )
		free( font_struct->per_char );
	DeleteObject(font_struct->fid);
	free(font_struct);
}

Bool FAR PASCAL
XGetFontProperty(XFontStruct *dummy1, Atom dummy2, 
	unsigned long *dummy3)
{
	return FALSE;
}

void FAR PASCAL
XTextExtents(XFontStruct *font_struct, const char *string, int nchars, 
	int *direction_return, int *ascent_return, int *descent_return, 
	XCharStruct *overall_return)
{
	HDC hdc;
	HFONT hfont_old;
	TEXTMETRIC tm;

	*direction_return = font_struct->direction;
	*ascent_return = font_struct->ascent;
	*descent_return = font_struct->descent;

	hdc = CreateCompatibleDC((HDC)NULL);
	hfont_old = SelectObject( hdc, font_struct->fid );

	GetTextMetrics( hdc, &tm );
	overall_return->ascent = tm.tmAscent;
	overall_return->descent = tm.tmDescent;
	overall_return->width = LOWORD(GetTextExtent( hdc, string, nchars ));
	overall_return->lbearing = 0;
	overall_return->rbearing = overall_return->width;

	SelectObject( hdc, hfont_old );
	DeleteDC( hdc );
}

int FAR PASCAL
XTextWidth(XFontStruct *font_struct, const char *string, int count)
{
	HDC hdc;
	HFONT hfont_old;
	int text_width;
	int one, two;

	hdc = GetDC((HWND)NULL);
	hfont_old = SelectObject( hdc, font_struct->fid );
	one = LOWORD(GetTextExtent( hdc, "a", 1 ));
	two = LOWORD(GetTextExtent( hdc, "aa", 2 ));
	text_width = LOWORD(GetTextExtent( hdc, string, count ));
	text_width = text_width - (one * 2 - two);
	SelectObject( hdc, hfont_old );
	ReleaseDC((HWND) NULL,  hdc );

	return text_width;
}
 
#ifdef KANJI

extern int decode(const wchar * ,char *, int);

int FAR PASCAL
XTextWidth16(XFontStruct *font_struct, const XChar2b * wstring, int count)
{
    wchar *string = (wchar *) wstring;
    int len;
    char *jp;
    int result;

    len = decode(string, (void *)NULL, count);
    jp = (char *)malloc(len + 1);
    (void) decode(string, jp, count);

    result = XTextWidth(font_struct, jp, len);
    free(jp);
    return result;
}
  
void FAR PASCAL
XTextExtents16(XFontStruct *font_struct, const XChar2b *string, int nchars,
	int *direction_return, int *ascent_return, int *descent_return,
	XCharStruct *overall_return)
{
    int len;
    char *jp;

    len = decode((wchar *)string, (void*)NULL, nchars);
    jp = (char *)malloc(len + 1);
    (void) decode((wchar *)string, jp, nchars);
    XTextExtents(font_struct, jp, len, direction_return, ascent_return,
	descent_return, overall_return);
    free(jp);
}
#endif /* KANJI */ 


