/*
 * xpropert.c --
 *
 * Copyright (c) 1994,1995 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include "twinint.h" 
#include <x11/xlib.h>
#include <x11/xutil.h>
#include <stdlib.h>
#include <string.h>

static void	NEAR PASCAL	_ChangeProperty(Window, char *, char *, int);

Status FAR PASCAL
XStringListToTextProperty(list, n, ptp)
char **list;
int n;
XTextProperty *ptp;
{
    ptp->value = (unsigned char *)malloc(strlen(*list) + 1);
    strcpy(ptp->value, *list);
    return 1;
}

void FAR PASCAL
XSetWMName(d, w, ptp)
Display *d;
Window w;
XTextProperty *ptp;
{
    _ChangeProperty(w, "XA_WM_NAME", ptp->value, strlen(ptp->value)+1);
    SetWindowText(w, ptp->value);
}

void FAR PASCAL
XSetIconName(d, w, p)
Display *d;
Window w;
const char *p;
{
    XTextProperty tp;
    
    XStringListToTextProperty((char **)&p, 1, &tp);
    _ChangeProperty(w, "XA_WM_ICON_NAME", tp.value, strlen(tp.value)+1);
    SendMessage(w, WM_SIZE, SIZE_MINIMIZED, 0L);
    XFree(tp.value);
}        	   

static void NEAR PASCAL
_ChangeProperty(w, np, dp, n)
Window w;
char *np, *dp;
int n;
{
    HGLOBAL hMem;
    char *p;
    if ((hMem = GetProp(w, np)) != NULL)
	GlobalFree(hMem);
    hMem = GlobalAlloc(GHND, n);
    p = (char *) GlobalLock(hMem);
    memcpy(p, dp, n);
    GlobalUnlock(hMem);
    SetProp(w, np, hMem);
}

typedef struct {
Atom type;
int format;
int nelem;
HGLOBAL hdata;
} PROP;

void FAR PASCAL
XChangeProperty(lpDisp, hWnd, atmProp, atmType, format, mode, lpData, nElem)
Display FAR * lpDisp;
HWND hWnd;
Atom atmProp, atmType;
int format;
int mode;
const unsigned char * lpData;
int nElem;
{
    HGLOBAL hGlobal, hw1;
    PROP FAR * lpprop;
    LPSTR lpw1, lpw2;

    if (hWnd == HWND_DESKTOP)
	hWnd = GetHomeWnd();

    if ((hGlobal = GetProp(hWnd, MAKELP(0, atmProp))) == NULL) {
    	hGlobal = GlobalAlloc(GHND, sizeof(PROP));
        mode = PropModeReplace;
    }
    if ((lpprop = (PROP FAR *) GlobalLock(hGlobal)) == NULL) {
	PostError(BadAlloc, hWnd);	
	return;
    }
    switch (mode) {
    case PropModeReplace:
labelPropModeReplace:
    	if (lpprop->hdata != 0)
	    GlobalFree(lpprop->hdata);
	lpprop->hdata = GlobalAlloc(GHND, (format / 8 * nElem));
	lpw1 = (LPSTR) GlobalLock(lpprop->hdata);
	_fmemcpy(lpw1, lpData, (format / 8 * nElem));
	GlobalUnlock(lpprop->hdata);
	lpprop->type = atmType;
	lpprop->format = format;
	lpprop->nelem = nElem;
	break;
    case PropModePrepend:
        if (lpprop->type != atmType || lpprop->format != format) {
	    PostError(BadMatch, hWnd);
	    return;
	}
    	if (lpprop->hdata != 0)
	    goto labelPropModeReplace;
	lpw1 = (LPSTR) GlobalLock(lpprop->hdata); 	
	hw1 = GlobalAlloc(GHND, format / 8 * (lpprop->nelem + nElem));
	lpw2 = (LPSTR) GlobalLock(hw1);
	_fmemcpy(lpw2, lpData, (format / 8 * nElem));
	_fmemcpy(lpw2 + (format / 8 * nElem), lpw1, (format / 8 * lpprop->nelem));
        GlobalUnlock(hw1);
        GlobalUnlock(lpprop->hdata);
	GlobalFree(lpprop->hdata);
	lpprop->hdata = hw1;
	lpprop->nelem += nElem;
	break;	   
    case PropModeAppend:
        if (lpprop->type != atmType || lpprop->format != format) {
	    PostError(BadMatch, hWnd);
	    return;
	}
    	if (lpprop->hdata != 0)
	    goto labelPropModeReplace;
	lpprop->hdata = GlobalReAlloc(lpprop->hdata, GHND, format / 8 * (lpprop->nelem + nElem));
	lpw1 = GlobalLock(lpprop->hdata);
	_fmemcpy(lpw1 + (format / 8 * lpprop->nelem), lpData, (format / 8 * nElem));
        GlobalUnlock(lpprop->hdata);
	lpprop->nelem += nElem;
	break;	   
    }
    GlobalUnlock(hGlobal);
    PostMessage(hWnd, PROPERTYNOTIFY, atmProp, PropertyNewValue);
    SetProp(hWnd, MAKELP(0, atmProp), hGlobal);
}

void FAR PASCAL
XDeleteProperty(lpDisp, hWnd, atmProp)
Display FAR * lpDisp;
HWND hWnd;
Atom atmProp;
{
    HGLOBAL hGlobal;
    PROP FAR * lpprop;

    if (hWnd == HWND_DESKTOP)
	hWnd = GetHomeWnd();

    if ((hGlobal = GetProp(hWnd, MAKELP(0, atmProp))) == NULL) {
	return; 
    }
    lpprop = (PROP FAR *) GlobalLock(hGlobal);
    if (lpprop->hdata != NULL) 
    	GlobalFree(lpprop->hdata);
    GlobalUnlock(hGlobal);
    GlobalFree(hGlobal);
    RemoveProp(hWnd, MAKELP(0, atmProp));
}

int FAR PASCAL
XGetWindowProperty(lpDisp, hWnd, atmProp, lOffset, lLemgth, delete, atmReqType, 
	lpatmActType, lpActFormat, lpnItems, lplByteAfter, lppData)
Display FAR * lpDisp;
HWND hWnd;
Atom atmProp;
long lOffset, lLemgth;
Bool delete;
Atom atmReqType;
Atom FAR * lpatmActType;
int FAR * lpActFormat;
unsigned long FAR * lpnItems;
unsigned long FAR * lplByteAfter;
unsigned char FAR * FAR * lppData;
{
    HGLOBAL hGlobal;
    PROP FAR * lpprop;
    LPSTR lpw1;

    if (hWnd == HWND_DESKTOP)
	hWnd = GetHomeWnd();

    if ((hGlobal = GetProp(hWnd, MAKELP(0, atmProp))) == NULL) {
not_exist:
	*lplByteAfter = 0;
	return 0;
    }
    if ((lpprop = (PROP FAR *) GlobalLock(hGlobal)) == NULL) {
    	RemoveProp(hWnd, MAKELP(0, atmProp));
	goto not_exist;
    }
    *lpatmActType = lpprop->type;
    *lpActFormat = lpprop->format;
    *lpnItems = lpprop->nelem;
    *lplByteAfter = 0;
    lpw1 = (LPSTR) GlobalLock(lpprop->hdata);
    *lppData = malloc(lpprop->format / 8 * lpprop->nelem);
    _fmemcpy(*lppData, lpw1, lpprop->format / 8 * lpprop->nelem);
    GlobalUnlock(lpprop->hdata); 
    GlobalUnlock(hGlobal); 

    return Success;
}
